Package: theft
Type: Package
Title: Tools for Handling Extraction of Features from Time Series
Version: 0.8.4
Date: 2026-04-29
Authors@R: c(
  person("Trent", "Henderson", email = "then6675@uni.sydney.edu.au", role = c("cre", "aut")),
  person("Annie", "Bryant", role = "ctb")
  )
Maintainer: Trent Henderson <then6675@uni.sydney.edu.au>
Description: Consolidates and calculates different sets of time-series features from multiple
    'R' and 'Python' packages including 'Rcatch22' Henderson, T. (2021) <doi:10.5281/zenodo.5546815>,
    'feasts' O'Hara-Wild, M., Hyndman, R., and Wang, E. (2021) <https://CRAN.R-project.org/package=feasts>,
    'tsfeatures' Hyndman, R., Kang, Y., Montero-Manso, P., Talagala, T., Wang, E., Yang, Y., and O'Hara-Wild, M. (2020)
    <https://CRAN.R-project.org/package=tsfeatures>, 'tsfresh' Christ, M., Braun, N., Neuffer, J.,
    and Kempa-Liehr A.W. (2018) <doi:10.1016/j.neucom.2018.03.067>, 'TSFEL' Barandas, M., et al. (2020)
    <doi:10.1016/j.softx.2020.100456>, and 'Kats' Facebook Infrastructure Data Science (2021)
    <https://facebookresearch.github.io/Kats/>.
BugReports: https://github.com/hendersontrent/theft/issues
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 3.5.0)
Imports: utils, stats, rlang, dplyr, tidyr, purrr, tsibble, fabletools,
        feasts, tsfeatures, Rcatch22, reticulate, R.matlab
Suggests: lifecycle, cachem, bslib, knitr, rmarkdown, pkgdown, testthat
RoxygenNote: 7.3.3
VignetteBuilder: knitr
URL: https://hendersontrent.github.io/theft/
NeedsCompilation: no
Packaged: 2026-04-29 02:34:36 UTC; trenthenderson
Author: Trent Henderson [cre, aut],
  Annie Bryant [ctb]
Repository: CRAN
Date/Publication: 2026-04-29 06:40:07 UTC
Built: R 4.4.3; ; 2026-05-01 15:21:17 UTC; unix

# Generated by roxygen2: do not edit by hand

export(calculate_features)
export(init_theft)
export(init_theft_hctsa)
export(init_theft_kats)
export(init_theft_tsfel)
export(init_theft_tsfresh)
export(install_kats)
export(install_pyhctsa)
export(install_python_pkgs)
export(install_tsfel)
export(install_tsfresh)
export(moments)
export(process_hctsa_file)
export(quantiles)
import(dplyr)
import(feasts)
importFrom(R.matlab,readMat)
importFrom(Rcatch22,catch22_all)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,distinct)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,reframe)
importFrom(dplyr,rename)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,sym)
importFrom(dplyr,ungroup)
importFrom(fabletools,feature_set)
importFrom(fabletools,features)
importFrom(purrr,map_df)
importFrom(reticulate,import)
importFrom(reticulate,source_python)
importFrom(reticulate,use_virtualenv)
importFrom(reticulate,virtualenv_create)
importFrom(reticulate,virtualenv_install)
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,ts)
importFrom(stats,var)
importFrom(tidyr,gather)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,unnest_wider)
importFrom(tsfeatures,tsfeatures)
importFrom(tsibble,as_tsibble)
importFrom(tsibble,index_var)
importFrom(tsibble,key_vars)
importFrom(utils,packageVersion)

# theft <img src="man/figures/logo.png" align="right" width="120" />

[![CRAN
version](https://www.r-pkg.org/badges/version/theft)](https://www.r-pkg.org/pkg/theft)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/theft)](https://www.r-pkg.org/pkg/theft)
[![DOI](https://zenodo.org/badge/351259952.svg)](https://zenodo.org/badge/latestdoi/351259952)

Tools for Handling Extraction of Features from Time series (theft)

## Installation

You can install the stable version of `theft` from CRAN:

``` r
install.packages("theft")
```

You can install the development version of `theft` from GitHub using the
following:

``` r
devtools::install_github("hendersontrent/theft")
```

Please also check out our paper [“Feature-Based Time-Series Analysis in
R using the Theft
Ecosystem”](https://journal.r-project.org/articles/RJ-2025-023/) which
discusses the motivation and theoretical underpinnings of `theft` and
walks through all of its functionality using the [Bonn EEG
dataset](https://pubmed.ncbi.nlm.nih.gov/11736210/) — a well-studied
neuroscience dataset.

## General purpose

`theft` is a software package for R that facilitates user-friendly
access to a consistent interface for the extraction of time-series
features. The package provides a single point of access to
$\approx 6000$ time-series features from a range of existing R and
Python packages as well as enabling users to calculate their own
features. The packages which `theft` ‘steals’ features from currently
are:

- [catch22](https://link.springer.com/article/10.1007/s10618-019-00647-x)
  (R; [see `Rcatch22` for the native implementation on
  CRAN](https://github.com/hendersontrent/Rcatch22))
- [feasts](https://feasts.tidyverts.org) (R)
- [tsfeatures](https://github.com/robjhyndman/tsfeatures) (R)
- [Kats](https://facebookresearch.github.io/Kats/) (Python)
- [tsfresh](https://tsfresh.com) (Python)
- [TSFEL](https://tsfel.readthedocs.io/en/latest/) (Python)
- [hctsa](https://github.com/DynamicsAndNeuralSystems/pyhctsa) (Python;
  originally MATLAB)

As of `v0.6.1`, users can also calculate their own individual features
or sets of features too! In addition, two basic feature sets
`"quantiles"` (a set of 101 quantiles) and `"moments"` (the first four
moments of the distribution: mean, variance, skewness, and kurtosis) are
also available for users seeking to compute simple baselines against
which to compare the more sophisticated feature sets (see [this recent
paper](https://arxiv.org/abs/2303.17809) for more discussion on this
idea).

Note that `Kats`, `tsfresh`, `TSFEL`, and `pyhctsa` are Python packages.
`theft` has built-in functionality for helping you install these
libraries—all you need to do is install Python on your machine
(preferably Python \>=3.10). If you wish to access the Python feature
sets, please run `?install_python_pkgs` in R after downloading `theft`
or consult the vignette in the package for more information. For a
comprehensive comparison of these six feature sets across a range of
domains (including computation speed, within-set feature composition,
and between-set feature correlations), please refer to the paper [An
Empirical Evaluation of Time-Series Feature
Sets](https://ieeexplore.ieee.org/document/9679937).

## Package extensibility

The companion package
[`theftdlc`](https://github.com/hendersontrent/theftdlc) (‘`theft`
downloadable content’—just like you get [DLCs and
expansions](https://en.bandainamcoent.eu/elden-ring/elden-ring/shadow-of-the-erdtree)
for video games) contains an extensive suite of functions for analysing,
interpreting, and visualising time-series features calculated from
`theft`. Collectively, these packages are referred to as the ‘`theft`
ecosystem’.

<img src="man/figures/theft-packages.png" alt="Hex stickers of the theft and theftdlc packages for R" />

A high-level overview of how the `theft` ecosystem for R is typically
accessed by users is shown below. Note that prior to `v0.6.1` of, many
of the `theftdlc` functions were contained in `theft` but under other
names. To ensure the `theft` ecosystem is as user-friendly as possible
and can scale to meet future demands, `theft` has been refactored to
just perform feature extraction, while `theftdlc` handles all the
processing, analysis, and visualisation of the extracted features.

<img src="man/figures/workflow-graphic_ecosystem-final.png" alt="Schematic of the theft ecosystem in R" />

Many more functions and options for customisation are available within
the packages and users are encouraged to explore the vignettes and
helper files for more information.

## Quick tour

`theft` and `theftdlc` combine to create an intuitive and efficient
workflow consistent with the broader
[`tidyverts`](https://tidyverts.org) collection of packages for tidy
time-series analysis. Here is a single code chunk that calculates
features for a [`tsibble`](https://tsibble.tidyverts.org) (tidy temporal
data frame) of some simulated time series processes, including Gaussian
noise, AR(1), ARMA(1,1), MA(1), noisy sinusoid, and a random walk.
`simData` comes with `theft`. We’ll just use the
[`catch22`](https://github.com/hendersontrent/Rcatch22) feature set and
a custom set of mean and standard deviation for now. Using tidy
principles and pipes, we can, in the same code chunk, feed the
calculated features straight into `theftdlc`’s `project` function to
project the 24-dimensional feature space into an interpretable
two-dimensional space using principal components analysis:

``` r
library(dplyr)
library(theft)
library(theftdlc)

calculate_features(data = theft::simData, 
                   feature_set = "catch22",
                   features = list("mean" = mean, "sd" = sd)) %>%
  project(norm_method = "RobustSigmoid",
          unit_int = TRUE,
          low_dim_method = "PCA") %>%
  plot()
```

![](README_files/figure-gfm/unnamed-chunk-4-1.png)<!-- -->

In that example, `calculate_features` comes from `theft`, while
`project` and the `plot` generic come from `theftdlc`.

Similarly, we can perform time-series classification using a similar
workflow to compare the performance of `catch22` against our custom set
of the first two moments of the distribution:

``` r
calculate_features(data = theft::simData, 
                   feature_set = "catch22",
                   features = list("mean" = mean, "sd" = sd)) %>%
  classify(by_set = TRUE,
           n_resamples = 10,
           use_null = TRUE) %>%
  compare_features(by_set = TRUE,
                   hypothesis = "pairwise") %>%
  head()
```

                   hypothesis feature_set_a feature_set_b   metric set_a_mean
    1 All features != catch22  All features       catch22 accuracy  0.8022222
    2    All features != User  All features          User accuracy  0.8022222
    3         catch22 != User       catch22          User accuracy  0.7400000
      set_b_mean t_statistic    p.value
    1  0.7400000  2.35154855 0.04319536
    2  0.8044444 -0.03932757 0.96948780
    3  0.8044444 -1.23794041 0.24705786

In this example, `classify` and `compare_features` come from `theftdlc`.

We can also easily see how each set performs relative to an empirical
null distribution (i.e., how much better does each set do than we would
expect due to chance?):

``` r
calculate_features(data = theft::simData, 
                   feature_set = "catch22",
                   features = list("mean" = mean, "sd" = sd)) %>%
  classify(by_set = TRUE,
           n_resamples = 10,
           use_null = TRUE) %>%
  compare_features(by_set = TRUE,
                   hypothesis = "null") %>%
  head()
```

                    hypothesis  feature_set   metric  set_mean null_mean
    1 All features != own null All features accuracy 0.8022222 0.1355556
    2         User != own null         User accuracy 0.8044444 0.1511111
    3      catch22 != own null      catch22 accuracy 0.7400000 0.1222222
      t_statistic      p.value
    1    6.826807 3.835233e-05
    2    5.882092 1.171092e-04
    3    6.879652 3.614676e-05

Please see the vignette for more information and the full functionality
of both packages.

## Citation

If you use `theft` or `theftdlc` in your own work, please cite both the
paper:

T. Henderson and Ben D. Fulcher. [“Feature-Based Time-Series Analysis in
R using the Theft
Ecosystem”](https://journal.r-project.org/articles/RJ-2025-023/), The R
Journal, 2025.

BibTeX version:

    @article{RJ-2025-023,
      author = {Henderson, Trent and Fulcher, Ben D.},
      title = {Feature-Based Time-Series Analysis in R using the Theft Ecosystem},
      journal = {The R Journal},
      year = {2025},
      note = {https://doi.org/10.32614/RJ-2025-023},
      doi = {10.32614/RJ-2025-023},
      volume = {17},
      issue = {3},
      issn = {2073-4859},
      pages = {43-68}
    }

and the software:

    To cite package 'theft' in publications use:

      Henderson T (2025). _theft: Tools for Handling Extraction of Features
      from Time Series_. R package version 0.8.3,
      <https://hendersontrent.github.io/theft/>.

    A BibTeX entry for LaTeX users is

      @Manual{,
        title = {theft: Tools for Handling Extraction of Features from Time Series},
        author = {Trent Henderson},
        year = {2025},
        note = {R package version 0.8.3},
        url = {https://hendersontrent.github.io/theft/},
      }

    To cite package 'theftdlc' in publications use:

      Henderson T (2025). _theftdlc: Analyse and Interpret Time Series
      Features_. R package version 0.2.1,
      <https://hendersontrent.github.io/theftdlc/>.

    A BibTeX entry for LaTeX users is

      @Manual{,
        title = {theftdlc: Analyse and Interpret Time Series Features},
        author = {Trent Henderson},
        year = {2025},
        note = {R package version 0.2.1},
        url = {https://hendersontrent.github.io/theftdlc/},
      }

## Acknowledgements

Big thanks to [Joshua Moore](https://github.com/joshuabmoore) for his
assistance in solving issues with the Python side of things and for
developing `pyhctsa`.

#' Calculate a basic set of the four moments of the distribution
#'
#' @importFrom stats var
#'
#' @param y \code{numeric} vector of values
#' @return \code{data.frame} of results
#' @author Trent Henderson
#' @export
#'

moments <- function(y){
  moments_df <- data.frame(
    names = c("mean", "variance", "skewness", "kurtosis"),
    values = c(mean(y), stats::var(y), skewness(y), kurtosis(y))
  )
  
  moments_df$feature_set <- "moments"
  return(moments_df)
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage("This is version ", utils::packageVersion(pkgname), 
                        " of ", pkgname, ". All analysis and visualisation functions (plus a host of new ones!) have been moved to the {theftdlc} package\nand subsequently renamed and refactored. You can install {theftdlc} via install.packages('theftdlc').\nPlease see https://hendersontrent.github.io/theftdlc for more.\ntheft also now requires a `tsibble` as an input rather than a data frame. Please consult the new documentation.")
}

#' Calculate a basic set of quantiles for an input time-series vector
#'
#' @importFrom stats quantile
#'
#' @param y \code{numeric} vector of values
#' @param quantiles \code{numeric} vector of quantiles to calculate. Defaults to \code{seq(0, 1, by = 0.01)}
#' @return \code{data.frame} of results
#' @author Trent Henderson
#' @export
#'

quantiles <- function(y, quantiles = seq(0, 1, by = 0.01)){
  quantiles_df <- data.frame(
    names = paste0("quantile_", quantiles * 100),
    values = sapply(quantiles, function(q) as.numeric(stats::quantile(y, q)))
  )
  
  quantiles_df$feature_set <- "quantiles"
  return(quantiles_df)
}

#---------------- Helper function ------------------

hctsa_iterator <- function(labels, keywords, timeSeriesData, x){
  
  tmp <- timeSeriesData[[x]][[1]]
  tmp <- as.data.frame(tmp)
  
  tmp <- tmp %>%
    dplyr::rename(values = .data$V1) %>%
    dplyr::mutate(timepoint = dplyr::row_number()) %>%
    dplyr::mutate(group = keywords[x],
                  id = labels[x])
  
  return(tmp)
}

#---------------- Core function ------------------

#' Load in hctsa formatted MATLAB files of time series data into a tidy format ready for feature extraction
#' 
#' @importFrom rlang .data
#' @importFrom R.matlab readMat
#' @importFrom purrr map_df
#' @import dplyr
#' @param data \code{string} specifying the filepath to the \code{MATLAB} file to parse
#' @return an object of class \code{data.frame} in tidy format
#' @author Trent Henderson
#' @export
#' 

process_hctsa_file <- function(data){
  
  if(!is.character(data)){
    stop("data should be a string specifying the filepath to the MATLAB file to parse")
  }
  
  # Read MATLAB file into list
  
  d <- R.matlab::readMat(data)
  theNames <- names(d)
  correctNames <- c("timeSeriesData", "labels", "keywords")
  
  if(length(theNames) != 3){
    stop("3 variables should be 'timeSeriesData', 'labels', and 'keywords'.")
  }
  
  if(setequal(theNames, correctNames) == FALSE){
    stop("3 variables should be 'timeSeriesData', 'labels', and 'keywords'.")
  }
  
  # Parse into single tidy dataframe
  
  keywords <- as.vector(do.call(rbind, d$keywords))
  labels <- as.vector(do.call(rbind, d$labels))
  timeSeriesData <- d$timeSeriesData
  
  myData <- 1:length(timeSeriesData) %>%
    purrr::map_df(~ hctsa_iterator(labels, keywords, timeSeriesData, .x))
  
  if(!is.numeric(myData$values)){
    stop("Non-numerics identified in values.")
  }
  
  return(myData)
}

#' Communicate to R the Python virtual environment containing the relevant libraries for calculating features
#' 
#' @importFrom reticulate use_virtualenv import
#'
#' @param venv \code{character} specifying the name of the to the Python virtual environment where \code{"tsfresh"}, \code{"TSFEL"}, and/or \code{"Kats"} are installed
#' @param hctsa \code{Boolean} denoting whether pyhctsa was installed or not. Defaults to \code{FALSE}
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test")
#' init_theft("theft-test")
#' }
#' 

init_theft <- function(venv, hctsa = FALSE){
  reticulate::use_virtualenv(venv)
  tsfresh <- reticulate::import("tsfresh")
  tsfel <- reticulate::import("tsfel")
  kats <- reticulate::import("kats")
  
  if(hctsa){
    pyhctsa <- reticulate::import("pyhctsa")
  }
}

#' Communicate to R the Python virtual environment containing tsfresh only
#' 
#' @importFrom reticulate use_virtualenv import
#'
#' @param venv \code{character} specifying the name of the to the Python virtual environment where \code{"tsfresh"} is installed
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test")
#' init_theft_tsfresh("theft-test")
#' }
#' 

init_theft_tsfresh <- function(venv){
  reticulate::use_virtualenv(venv)
  tsfresh <- reticulate::import("tsfresh")
}

#' Communicate to R the Python virtual environment containing tsfel only
#' 
#' @importFrom reticulate use_virtualenv import
#'
#' @param venv \code{character} specifying the name of the to the Python virtual environment where \code{"tsfel"} is installed
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test")
#' init_theft_tsfel("theft-test")
#' }
#' 

init_theft_tsfel <- function(venv){
  reticulate::use_virtualenv(venv)
  tsfel <- reticulate::import("tsfel")
}

#' Communicate to R the Python virtual environment containing kats only
#' 
#' @importFrom reticulate use_virtualenv import
#'
#' @param venv \code{character} specifying the name of the to the Python virtual environment where \code{"kats"} is installed
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test")
#' init_theft_kats("theft-test")
#' }
#' 

init_theft_kats <- function(venv){
  reticulate::use_virtualenv(venv)
  kats <- reticulate::import("kats")
}

#' Communicate to R the Python virtual environment containing pyhctsa only
#' 
#' @importFrom reticulate use_virtualenv import
#'
#' @param venv \code{character} specifying the name of the to the Python virtual environment where \code{"pyhctsa"} is installed
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test")
#' init_theft_hctsa("theft-test")
#' }
#' 

init_theft_hctsa <- function(venv){
  reticulate::use_virtualenv(venv)
  pyhctsa <- reticulate::import("pyhctsa")
}

#' Check for presence of NAs and non-numerics in a vector
#' 
#' @param x input \code{vector}
#' @return \code{Boolean} of whether the data is good to extract features on or not
#' @author Trent Henderson
#' 

check_vector_quality <- function(x){
  return(!anyNA(x) && is.numeric(x))
}

#' Calculate a skewness of a vector
#'
#' @importFrom stats sd
#'
#' @param y \code{numeric} vector of values
#' @return \code{numeric} scalar of skewness
#' @author Trent Henderson
#'
skewness <- function(y){
  skew <- sum((y - mean(y))^3) / ((length(y) - 1) * stats::sd(y)^3)
  return(skew)
}

#' Calculate a kurtosis of a vector
#'
#' @importFrom stats sd
#'
#' @param y \code{numeric} vector of values
#' @return \code{numeric} scalar of kurtosis
#' @author Trent Henderson
#'
kurtosis <- function(y){
  kurt <- sum((y - mean(y))^4) / ((length(y) - 1) * stats::sd(y)^4)
  return(kurt)
}

#------------------- Helper functions to reduce length -------------

#--------
# catch22
#--------

calc_catch22 <- function(data, catch24){
  
  outData <- data %>%
    dplyr::reframe(Rcatch22::catch22_all(!!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]), 
                                         catch24 = catch24), .by = tsibble::key_vars(data)) %>%
    dplyr::mutate(feature_set = "catch22")
  
  return(outData)
}

#-------
# feasts
#-------

calc_feasts <- function(data){
  
  outData <- data %>%
    fabletools::features(!!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]),
                         features = fabletools::feature_set(pkgs = "feasts")) %>%
    tidyr::gather("names", "values", -dplyr::all_of(tsibble::key_vars(data))) %>%
    dplyr::mutate(feature_set = "feasts")
  
  return(outData)
}

#-----------
# tsfeatures
#-----------

calc_tsfeatures <- function(data, use_compengine, n_jobs){
  
  if(n_jobs >= 2){
    parallel <- TRUE
  } else{
    parallel <- FALSE
  }
  
  var1 <- tsibble::index_var(data)
  var2 <- colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]
  var3 <- tsibble::key_vars(data)[1]
  tsf_list <- split(data[, c(var2)], data[, var3])
  
  if(length(tsibble::key_vars(data)) > 1){
    lookup <- data %>%
      as.data.frame() %>%
      dplyr::select(dplyr::all_of(tsibble::key_vars(data))) %>%
      dplyr::distinct()
  }
  
  outData <- lapply(tsf_list, function(x){
    stats::ts(x)
  })
  
  the_names <- data.frame(id = names(outData))
  colnames(the_names) <- var3
  
  featureList <- c("frequency", "stl_features", "entropy", "acf_features",
                   "arch_stat", "crossing_points", "flat_spots",
                   "heterogeneity", "holt_parameters", "hurst", 
                   "lumpiness", "max_kl_shift", "max_level_shift", "max_var_shift", 
                   "nonlinearity", "pacf_features", "stability", "unitroot_kpss",
                   "unitroot_pp", "embed2_incircle", "firstzero_ac",
                   "histogram_mode", "localsimple_taures", "sampenc",
                   "spreadrandomlocal_meantaul")
  
  if(use_compengine){
    featureList <- append(featureList, "compengine")
  }
  
  outData <- try(tsfeatures::tsfeatures(outData, features = featureList, parallel = parallel))
  
  if("try-error" %in% class(outData)){
    message("Removing 'compengine' features from tsfeatures due to error. Recomputing with reduced set...")
    featureList <- featureList[!featureList %in% c("compengine")]
    outData <- tsfeatures::tsfeatures(outData, features = featureList, parallel = parallel)
  }
  
  outData <- cbind(the_names, outData) %>%
    tidyr::pivot_longer(!dplyr::all_of(var3), names_to = "names", values_to = "values") %>%
    dplyr::mutate(feature_set = "tsfeatures")
  
  if(length(tsibble::key_vars(data)) > 1){
    outData <- outData %>%
      dplyr::inner_join(lookup)
  }
  
  return(outData)
}

#--------
# tsfresh
#--------

calc_tsfresh <- function(data, cleanup, n_jobs = 0, warn = TRUE){
  
  mywarn <- ifelse(warn, "Yes", "No")
  
  lookup <- data %>%
    as.data.frame() %>%
    dplyr::select(dplyr::all_of(tsibble::key_vars(data))) %>%
    dplyr::distinct()
  
  # Load Python function
  
  tsfresh_calculator <- function(){}
  reticulate::source_python(system.file("python", "tsfresh_calculator.py", package = "theft")) # Ships with package
    
  temp <- data %>%
    as.data.frame() %>%
    dplyr::group_by(!!dplyr::sym(tsibble::key_vars(data)[1])) %>%
    dplyr::arrange(!!dplyr::sym(tsibble::index_var(data))) %>%
    # dplyr::mutate(timepoint = as.numeric(dplyr::row_number())) %>%
    dplyr::ungroup() %>%
    dplyr::select(!!dplyr::sym(tsibble::key_vars(data)[1]), !!dplyr::sym(tsibble::index_var(data)),
                  !!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]))
  
  ids <- temp %>%
    dplyr::select(!!dplyr::sym(tsibble::key_vars(data)[1])) %>%
    dplyr::distinct()
    
  outData <- tsfresh_calculator(timeseries = temp, column_id = tsibble::key_vars(data)[1],
                                column_sort = tsibble::index_var(data),
                                cleanup = cleanup, n_jobs = n_jobs, warn = mywarn) %>%
    cbind(ids) %>%
    tidyr::gather("names", "values", -tsibble::key_vars(data)[1]) %>%
    dplyr::inner_join(lookup) %>%
    dplyr::mutate(feature_set = "tsfresh")
  
  return(outData)
}

#------
# TSFEL
#------

calc_tsfel <- function(data, n_jobs, warn){
  
  mywarn <- ifelse(warn, "Yes", "No")
  
  # Load Python function
  
  tsfel_calculator <- function(){}
  reticulate::source_python(system.file("python", "tsfel_calculator.py", package = "theft")) # Ships with package
  
  outData <- data %>%
    dplyr::reframe(tsfel_calculator(!!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]),
                                    n_jobs = n_jobs, warn = mywarn), 
                   .by = dplyr::all_of(tsibble::key_vars(data))) %>%
    tidyr::gather("names", "values", -tsibble::key_vars(data)) %>%
    dplyr::mutate(feature_set = "TSFEL")
  
  return(outData)
}

#-----
# Kats
#-----

calc_kats <- function(data, warn){
  
  mywarn <- ifelse(warn, "Yes", "No")
  
  # Load Python function
  
  kats_calculator <- function(){}
  reticulate::source_python(system.file("python", "kats_calculator.py", package = "theft")) # Ships with package
  
  # Convert numeric time index to datetime as Kats requires it
  
  unique_times <- data %>%
    as.data.frame() %>%
    dplyr::select(!!dplyr::sym(tsibble::index_var(data))) %>%
    dplyr::distinct() %>%
    dplyr::pull(!!dplyr::sym(tsibble::index_var(data)))
  
  datetimes <- data.frame(timepoint = unique_times) %>%
    dplyr::mutate(time = seq(as.Date("1800-01-01"), by = "day", length.out = length(unique_times)))
  
  colnames(datetimes) <- c(tsibble::index_var(data), "time")
  
  # Join in datetimes and run computations
  
  outData <- data %>%
    dplyr::inner_join(datetimes) %>%
    dplyr::select(-!!dplyr::sym(tsibble::index_var(data))) %>%
    dplyr::reframe(results = list(kats_calculator(timepoints = .data$time, 
                                                  values = !!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]),
                                                  warn = mywarn)), 
                   .by = dplyr::all_of(tsibble::key_vars(data))) %>%
    tidyr::unnest_wider(!!dplyr::sym("results")) %>%
    tidyr::gather("names", "values", -tsibble::key_vars(data)) %>%
    dplyr::mutate(feature_set = "Kats")
  
  return(outData)
}

#------
# hctsa
#------

calc_hctsa <- function(data, warn){
  
  mywarn <- ifelse(warn, "Yes", "No")
  
  # Load Python function
  
  pyhctsa_calculator <- function(){}
  reticulate::source_python(system.file("python", "pyhctsa_calculator.py", package = "theft")) # Ships with package
  hctsa_config <- system.file("yaml", "hctsa.yaml", package = "theft") # Ships with package
  
  # Wrangle data into matrix format for hctsa
  
  ts_list <- split(data$values, data$id)
  sel_cols <- c("id", "timepoint", "values")
  
  ts_list <- data %>%
    as.data.frame() %>%
    dplyr::select(dplyr::all_of(sel_cols)) %>%
    tidyr::pivot_wider(id_cols = "id", names_from = "timepoint", values_from = "values")
  
  idx <- ts_list$id
  
  ts_list <- ts_list %>%
    dplyr::select(-c("id")) %>%
    as.matrix()
  
  # Calculate features
  
  outData <- pyhctsa_calculator(ts_list, mywarn, hctsa_config)
  
  # Clean up data returns
  
  outData <- outData %>%
    dplyr::mutate(id = idx)
  
  if(length(tsibble::key_vars(data)) > 1){
    
    group_labs <- data %>%
      as.data.frame() %>%
      dplyr::select(dplyr::all_of(tsibble::key_vars(data))) %>%
      dplyr::distinct()
    
    outData <- outData %>%
      dplyr::inner_join(group_labs, by = c("id" = "id"))
  }
  
  outData <- outData %>%
    dplyr::select(c(dplyr::where(is.numeric), dplyr::all_of(tsibble::key_vars(data)))) %>%
    tidyr::pivot_longer(cols = !tsibble::key_vars(data), names_to = "names", values_to = "values") %>%
    dplyr::mutate(feature_set = "hctsa")
  
  return(outData)
}

#-----
# User
#-----

calc_user <- function(data, features){
  
  outData <- data %>%
    dplyr::reframe(dplyr::across(dplyr::all_of(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]), 
                                 .fns = features), 
                   .by = tsibble::key_vars(data))
  
  colnames(outData) <- append(tsibble::key_vars(data), names(features))
  key_var_count <- length(tsibble::key_vars(data)) + 1
  col_count <- ncol(outData)
  cols <- key_var_count:col_count
    
  outData <- outData %>%
    tidyr::pivot_longer(cols = dplyr::all_of(cols), names_to = "names", values_to = "values") %>%
    dplyr::mutate(feature_set = "User")
  
  return(outData)
}

#------------------- Main exported calculation function ------------

#' Compute features on an input time series dataset
#' 
#' @importFrom rlang .data :=
#' @importFrom dplyr group_by filter ungroup bind_rows across all_of select rename %>% mutate sym
#' @importFrom tsibble key_vars index_var
#' @param data \code{tbl_ts} containing the time series data
#' @param feature_set \code{character} or \code{vector} of \code{character} denoting the set of time-series features to calculate. Can be one of \code{"catch22"}, \code{"feasts"}, \code{"tsfeatures"}, \code{"tsfresh"}, \code{"tsfel"}, \code{"kats"}, \code{"quantiles"}, and or \code{"moments"}
#' @param features named \code{list} containing a set of user-supplied functions to calculate on \code{data}. Each function should take a single argument which is the time series. Defaults to \code{NULL} for no manually-specified features. Each list entry must have a name as \code{calculate_features} looks for these to name the features. If you don't want to use the existing feature sets and only compute those passed to \code{features}, set \code{feature_set = NULL}
#' @param catch24 \code{Boolean} specifying whether to compute \code{catch24} in addition to \code{catch22} if \code{catch22} is one of the feature sets selected. Defaults to \code{FALSE}
#' @param tsfresh_cleanup \code{Boolean} specifying whether to use the in-built \code{tsfresh} relevant feature filter or not. Defaults to \code{FALSE}
#' @param use_compengine \code{Boolean} specifying whether to use the \code{"compengine"} features in \code{tsfeatures}. Defaults to \code{FALSE} to provide immense computational efficiency benefits
#' @param seed \code{integer} denoting a fixed number for R's random number generator to ensure reproducibility. Defaults to \code{123}
#' @param z_score \code{Boolean} specifying whether to z-score the time-series before computing features. Defaults to \code{FALSE}
#' @param n_jobs \code{integer} denoting the number of parallel processes to use if \code{"tsfresh"} or \code{"tsfel"} are specified in \code{"feature_set"}. Defaults to \code{0} for no parallelisation
#' @param warn \code{Boolean} specifying whether to produce warnings from feature set packages. Defaults to \code{TRUE}
#' @return object of class \code{feature_calculations} that contains the summary statistics for each feature
#' @author Trent Henderson
#' @export
#' @examples
#' featMat <- calculate_features(data = simData, 
#'   feature_set = "catch22")
#'

calculate_features <- function(data, feature_set = c("catch22", "feasts", "tsfeatures", 
                                                     "kats", "tsfresh", "tsfel", "hctsa",
                                                     "quantiles", "moments"), 
                               features = NULL, catch24 = FALSE, 
                               tsfresh_cleanup = FALSE, use_compengine = FALSE, 
                               seed = 123, z_score = FALSE, n_jobs = 0, warn = TRUE){
  
  if(!inherits(data, "tbl_ts")){
    stop("As of v0.8.1 `data` must now be a `tbl_ts object`. Please convert your matrix or dataframe using `tsibble::as_tsibble` and specify your `key` and `index` variables.")
  }
  
  stopifnot(inherits(data, "tbl_ts"))
  
  if(ncol(data) > length(tsibble::key_vars(data)) + length(tsibble::index_var(data)) + 1){
    stop("Multiple measured variables detected. Please ensure there is only one measure variable in `data` outside of the `key` and `index` variables used to create the `tbl_ts` object.")
  }
  
  stopifnot(n_jobs >= 0)
  
  feature_set <- tolower(feature_set) # Standardise names
  
  #--------- Filter out time series with NAs --------
  
  ids_pre <- data %>%
    as.data.frame() %>%
    dplyr::select(dplyr::all_of(tsibble::key_vars(data)[1])) %>%
    dplyr::distinct() %>%
    nrow()
  
  data_re <- data %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(tsibble::key_vars(data)))) %>%
    dplyr::filter(!any(is.na(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]))) %>%
    dplyr::ungroup()
  
  lookup2 <- unique(data_re[tsibble::key_vars(data_re)])
  ids_post <- nrow(lookup2)
  
  if(ids_pre != ids_post){
    message(paste0("Removed ", ids_post - ids_pre, " time series due to non-real values."))
  }
  
  if(ids_post == 0){
    stop("No time series remaining to calculate features after removing IDs with non-real values.")
  }
  
  #--------- Normalise data --------
  
  if(z_score){
    data_re <- data_re %>%
      dplyr::group_by(dplyr::across(dplyr::all_of(tsibble::key_vars(data)))) %>%
      dplyr::mutate(!!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))]) := 
                      as.numeric(scale(!!dplyr::sym(colnames(data)[!colnames(data) %in% append(tsibble::key_vars(data), tsibble::index_var(data))])))) %>%
      dplyr::ungroup()
  }
  
  #--------- Feature calcs --------
  
  if("catch22" %in% feature_set){
      message("Running computations for catch22...\n")
    if(!warn){
      tmp_catch22 <-suppressWarnings(calc_catch22(data = data_re, catch24 = catch24))
    } else{
      tmp_catch22 <- calc_catch22(data = data_re, catch24 = catch24)
      }
    }
    
    if("feasts" %in% feature_set){
      message("Running computations for feasts...\n")
      if(!warn){
        tmp_feasts <- suppressWarnings(calc_feasts(data = data_re))
      } else{
        tmp_feasts <- calc_feasts(data = data_re)
      }
    }
    
    if("tsfeatures" %in% feature_set){
      message("Running computations for tsfeatures...\n")
      if(!warn){
        tmp_tsfeatures <- suppressWarnings(calc_tsfeatures(data = data_re, use_compengine = use_compengine, n_jobs = n_jobs))
      } else{
        tmp_tsfeatures <- calc_tsfeatures(data = data_re, use_compengine = use_compengine, n_jobs = n_jobs)
      }
    }
    
    if("tsfresh" %in% feature_set){
      
      if(tsfresh_cleanup){
        cleanuper <- "Yes"
      } else{
        cleanuper <- "No"
      }
      
      message("Running computations for tsfresh...\n")
      if(!warn){
        tmp_tsfresh <- suppressWarnings(calc_tsfresh(data = data_re, cleanup = cleanuper, n_jobs = as.integer(n_jobs), warn = warn))
      } else{
        tmp_tsfresh <- calc_tsfresh(data = data_re, cleanup = cleanuper, n_jobs = as.integer(n_jobs), warn = warn)
      }
    }
    
    if("tsfel" %in% feature_set){
      message("Running computations for TSFEL...\n")
      if(!warn){
        tmp_tsfel <- suppressWarnings(calc_tsfel(data = data_re, n_jobs = as.integer(n_jobs), warn = warn))
      } else{
        tmp_tsfel <- calc_tsfel(data = data_re, n_jobs = as.integer(n_jobs), warn = warn)
      }
    }
    
    if("kats" %in% feature_set){
      message("Running computations for Kats...\n")
      if(!warn){
        tmp_kats <- suppressWarnings(calc_kats(data = data_re, warn = warn))
      } else{
        tmp_kats <- calc_kats(data = data_re, warn = warn)
      }
    }
  
  if("hctsa" %in% feature_set){
    message("Running computations for hctsa...\n")
    if(!warn){
      tmp_hctsa <- suppressWarnings(calc_hctsa(data = data_re, warn = warn))
    } else{
      tmp_hctsa <- calc_hctsa(data = data_re, warn = warn)
    }
  }
    
    if("quantiles" %in% feature_set){
      message("Running computations for quantiles...\n")
      if(!warn){
        tmp_quantiles <- suppressWarnings(
          data_re %>%
            dplyr::reframe(quantiles(!!dplyr::sym(colnames(data_re)[!colnames(data_re) %in% append(tsibble::key_vars(data_re), tsibble::index_var(data_re))])), 
                           .by = tsibble::key_vars(data))
        )
      } else{
        tmp_quantiles <- data_re %>%
          dplyr::reframe(quantiles(!!dplyr::sym(colnames(data_re)[!colnames(data_re) %in% append(tsibble::key_vars(data_re), tsibble::index_var(data_re))])), 
                         .by = tsibble::key_vars(data))
      }
    }
    
    if("moments" %in% feature_set){
      message("Running computations for moments...\n")
      if(!warn){
        tmp_moments <- suppressWarnings(
          data_re %>%
            dplyr::reframe(moments(!!dplyr::sym(colnames(data_re)[!colnames(data_re) %in% append(tsibble::key_vars(data_re), tsibble::index_var(data_re))])), 
                           .by = tsibble::key_vars(data))
        )
      } else{
        tmp_moments <- data_re %>%
          dplyr::reframe(moments(!!dplyr::sym(colnames(data_re)[!colnames(data_re) %in% append(tsibble::key_vars(data_re), tsibble::index_var(data_re))])), 
                         .by = tsibble::key_vars(data))
      }
    }
    
    #-----------------------
    # User-supplied features
    #-----------------------
    
    if(!is.null(features)){
      stopifnot(class(features) == "list")
      stopifnot(sapply(features, class) == "function")
      
      if(is.null(names(features))){
        stop("features must be a named list as calculate_features uses the names to label features produced by each function in the list") # More informative error message than above as this is a bit more specific
      }
      
      message("Running computations for user-supplied features...\n")
      tmp_user <- calc_user(data = data_re, features = features)
    }
  
  #--------- Feature binding --------
  
  tmp_all_features <- data.frame()
  
  if(length(feature_set) > 1){
    message("Binding feature dataframes together...\n")
  }
  
  if(exists("tmp_catch22")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_catch22)
  }
  
  if(exists("tmp_feasts")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_feasts)
  }
  
  if(exists("tmp_tsfeatures")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_tsfeatures)
  }
  
  if(exists("tmp_tsfresh")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_tsfresh)
  }
  
  if(exists("tmp_tsfel")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_tsfel)
  }
  
  if(exists("tmp_kats")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_kats)
  }
  
  if(exists("tmp_hctsa")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_hctsa)
  }
  
  if(exists("tmp_quantiles")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_quantiles)
  }
  
  if(exists("tmp_moments")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_moments)
  }
  
  if(exists("tmp_user")){
    tmp_all_features <- dplyr::bind_rows(tmp_all_features, tmp_user)
  }
  
  # Change column names to be consistent with {theftdlc} package
  
  keep_cols <- c("id", "group", "feature_set", "names", "values")
  
  if(length(tsibble::key_vars(data)) > 1){
    tmp_all_features <- tmp_all_features %>%
      dplyr::rename(id = dplyr::all_of(tsibble::key_vars(data)[1]),
                    group = dplyr::all_of(tsibble::key_vars(data)[2])) %>%
      dplyr::select(dplyr::all_of(keep_cols))
  } else{
    if(tsibble::key_vars(data)[1] != "id"){
      tmp_all_features <- tmp_all_features %>%
        dplyr::rename(id = dplyr::all_of(tsibble::key_vars(data)[1]))
    }
  }
  
  tmp_all_features <- structure(tmp_all_features, class = c("feature_calculations", "data.frame"))
  return(tmp_all_features)
}

context("test-calculate_features")

test_that("catch22 feature calculation", {
  
  expect_equal(22 * length(unique(theft::simData$id)), nrow(feature_matrix))
})

test_that("feasts feature calculation", {
  
  skip_on_cran()
  
  features_feasts <- calculate_features(data = theft::simData, 
                                        feature_set = "feasts")
  
  expect_equal(43 * length(unique(theft::simData$id)), nrow(features_feasts))
})

test_that("custom feature calculation", {
  
  features_custom <- calculate_features(data = theft::simData, 
                                        feature_set = NULL, 
                                        features = list("mean" = mean, "sd" = sd))
  
  expect_equal(2 * length(unique(theft::simData$id)), nrow(features_custom))
})

#' All features available in theft in tidy format
#'
#' The variables include:
#'
#' @format A tidy data frame with 2 variables:
#' \describe{
#'   \item{feature_set}{Name of the set the feature is from}
#'   \item{feature}{Name of the feature}
#' }
#'
"feature_list"


#' Sample of randomly-generated time series to produce function tests and vignettes
#'
#' The variables include:
#'
#' @format A tidy tsibble with 4 variables:
#' \describe{
#'   \item{id}{Unique identifier for the time series}
#'   \item{timepoint}{Time index}
#'   \item{values}{Value}
#'   \item{process}{Group label for the type of time series}
#' }
#'
"simData"

"_PACKAGE"
#' @name theft
#' @title Tools for Handling Extraction of Features from Time-series
#' 
#' @description Tools for Handling Extraction of Features from Time-series
#' 
#' @importFrom utils packageVersion
#' @importFrom rlang .data
#' @importFrom stats ts
#' @importFrom dplyr %>% group_by filter ungroup bind_rows across all_of mutate arrange reframe mutate select distinct row_number left_join everything rename inner_join sym
#' @importFrom tsibble key_vars index_var as_tsibble
#' @importFrom fabletools features feature_set
#' @importFrom Rcatch22 catch22_all
#' @importFrom tsfeatures tsfeatures
#' @importFrom tidyr gather unnest_wider pivot_longer
#' @importFrom fabletools features feature_set
#' @importFrom R.matlab readMat
#' @importFrom reticulate virtualenv_create virtualenv_install use_virtualenv import source_python
#' @importFrom purrr map_df
#' @import feasts
NULL

#' Download and install tsfresh, TSFEL, and Kats from Python into a new virtual environment
#'
#' @importFrom reticulate virtualenv_create virtualenv_install
#' 
#' @param venv \code{character} specifying the name of the new virtual environment to create
#' @param python \code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended
#' @param hctsa \code{Boolean} denoting whether to install pyhctsa or not. Defaults to \code{FALSE}
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_python_pkgs("theft-test", "/usr/local/bin/python3.10")
#' }
#' 
install_python_pkgs <- function(venv, python, hctsa = FALSE){
  reticulate::virtualenv_create(venv, python)
  reticulate::virtualenv_install(venv, "tsfel")
  reticulate::virtualenv_install(venv, "tsfresh")
  reticulate::virtualenv_install(venv, "git+https://github.com/hendersontrent/theft-kats.git")
  
  if(hctsa){
    reticulate::virtualenv_install(venv, "git+https://github.com/DynamicsAndNeuralSystems/pyhctsa.git")
  }
  
  reticulate::virtualenv_install(venv, "scipy==1.14.0")
  reticulate::virtualenv_install(venv, "wheel==0.45.1")
  reticulate::virtualenv_install(venv, "packaging==21.3")
  reticulate::virtualenv_install(venv, "numpy==1.26")
}

#' Download and install tsfresh from Python into a new virtual environment
#'
#' @importFrom reticulate virtualenv_create virtualenv_install
#' 
#' @param venv \code{character} specifying the name of the new virtual environment to create
#' @param python \code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_tsfresh("theft-test", "/usr/local/bin/python3.10")
#' }
#' 
install_tsfresh <- function(venv, python){
  reticulate::virtualenv_create(venv, python)
  reticulate::virtualenv_install(venv, "tsfresh")
  reticulate::virtualenv_install(venv, "scipy==1.14.0")
  reticulate::virtualenv_install(venv, "wheel==0.45.1")
  reticulate::virtualenv_install(venv, "packaging==21.3")
  reticulate::virtualenv_install(venv, "numpy==1.26")
}

#' Download and install TSFEL from Python into a new virtual environment
#'
#' @importFrom reticulate virtualenv_create virtualenv_install
#' 
#' @param venv \code{character} specifying the name of the new virtual environment to create
#' @param python \code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_tsfel("theft-test", "/usr/local/bin/python3.10")
#' }
#' 
install_tsfel <- function(venv, python){
  reticulate::virtualenv_create(venv, python)
  reticulate::virtualenv_install(venv, "tsfel")
  reticulate::virtualenv_install(venv, "scipy==1.14.0")
  reticulate::virtualenv_install(venv, "wheel==0.45.1")
  reticulate::virtualenv_install(venv, "packaging==21.3")
  reticulate::virtualenv_install(venv, "numpy==1.26")
}

#' Download and install Kats from Python into a new virtual environment
#'
#' @importFrom reticulate virtualenv_create virtualenv_install
#' 
#' @param venv \code{character} specifying the name of the new virtual environment to create
#' @param python \code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_kats("theft-test", "/usr/local/bin/python3.10")
#' }
#' 
install_kats <- function(venv, python){
  reticulate::virtualenv_create(venv, python)
  reticulate::virtualenv_install(venv, "git+https://github.com/hendersontrent/theft-kats.git")
  reticulate::virtualenv_install(venv, "scipy==1.14.0")
  reticulate::virtualenv_install(venv, "wheel==0.45.1")
  reticulate::virtualenv_install(venv, "packaging==21.3")
  reticulate::virtualenv_install(venv, "numpy==1.26")
}

#' Download and install pyhctsa from Python into a new virtual environment
#'
#' @importFrom reticulate virtualenv_create virtualenv_install
#' 
#' @param venv \code{character} specifying the name of the new virtual environment to create
#' @param python \code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended
#' @return no return value; called for side effects
#' @author Trent Henderson
#' @export
#' @examples
#' \dontrun{
#' install_pyhctsa("theft-test", "/usr/local/bin/python3.10")
#' }
#' 
install_pyhctsa <- function(venv, python){
  reticulate::virtualenv_create(venv, python)
  reticulate::virtualenv_install(venv, "git+https://github.com/DynamicsAndNeuralSystems/pyhctsa.git")
}

---
title: "Introduction to theft"
author: "Trent Henderson"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
vignette: >
  %\VignetteIndexEntry{Introduction to theft}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.height = 7,
  fig.width = 7,
  warning = FALSE,
  fig.align = "center"
)
```

```{r setup, message = FALSE, warning = FALSE}
library(theft)
```

## Purpose

`theft` enables the standardised calculation of time-series features from seven existing feature sets in both R and Python as well as any user-supplied features.

## Core functionality

All time-series datasets passed into `theft` must be a `tbl_ts` generated through the [`tsibble`](https://tsibble.tidyverts.org/) R package. This ensures consistency with the broader [`tidyverts`](https://tidyverts.org/) collection of packages. To explore package functionality, we are going to use a dataset that comes with `theft` called `simData`. This dataset contains some simulated time series processes, including Gaussian noise, AR(1), ARMA(1,1), MA(1), noisy sinusoid, and a random walk. The dataset can be accessed via:
  
```{r, message = FALSE, warning = FALSE, eval = FALSE}
theft::simData
```

Note that `simData` is a `tsibble` with two `key` variables: `id` and `process` which identify each time series by its unique ID and group, and an `index` variable of `timepoint` which denotes time indices.

The data follows the following structure:
  
```{r, message = FALSE, warning = FALSE}
head(simData)
```

### Calculating time-series features

The core function in `theft` is `calculate_features`. You can choose which subset of features to calculate with the `feature_set` argument. The choices are currently `"catch22"`, `"feasts"`, `"tsfeatures"`, `"tsfresh"`, `"tsfel"`, `"kats"`, and `"hctsa"`. In addition, two basic feature sets `"quantiles"` (a set of 101 quantiles) and `"moments"` (the first four moments of the distribution: mean, variance, skewness, and kurtosis) are also available to be specified for users seeking to compute simple baselines against which to compare the more sophisticated feature sets (see [this recent paper](https://arxiv.org/abs/2303.17809) for more discussion on this idea).

Note that `kats`, `tsfresh`, `tsfel`, and `pyhctsa` are Python packages. The R package `reticulate` is used to call Python code that uses these packages and applies it within the broader *tidy* data philosophy embodied by `theft`. `theft` currently provides access to $\approx 6000$ features from these seven sets alone. However, as discussed in the functionality demonstrations below, you can also supply your own list of features too!

#### Installing Python feature sets

Prior to using `theft` (only if you want to use the `kats`, `tsfresh`, `tsfel`, or `hctsa` feature sets; the R-based sets will run fine) you should have a working Python 3.9 installation and run the function `install_python_pkgs(venv, python, hctsa = FALSE)` (or with `hctsa = TRUE` if you want that large feature set) after first installing `theft`, where the `venv` argument is the name of the virtual environment you want to create and `python` is the path to the Python interpreter you want to use. Note that the filepath to the Python interpreter will differ depending on your operating system. For example, on Windows it might be a path similar to `"C:/Users/YourName/AppData/Local/Programs/Python/Python310/python.exe"`, whereas on on Linux or MacOS (which was used for this vignette) it could be a path similar to `"/usr/local/bin/python3.10"` or `"/usr/bin/python3"`.

For example, if you wanted to install the Python libraries to the default virtual environment folder used by `reticulate`, you would run the following after first having installed `theft` (here I am just creating a new virtual environment called `"theft-package"`---you can call it whatever you like!):

```{r, eval = FALSE}
install_python_pkgs(venv = "theft-package", python = "/usr/local/bin/python3.10")
```

You can then run the following to activate the virtual environment:

```{r, eval = FALSE}
init_theft("theft-package")
```

You are now ready to commit theft!

**NOTE 1: You only need to call ** `init_theft` **once per session.**

**NOTE 2: There are also separate installation functions for each Python feature set, such as** `install_tsfresh` **if you only need one of the libraries and want to keep your dependencies light.**

**NOTE 3: There are also separate** `init_theft` **functions for if you are interested in only using one of the Python feature sets, such as** `init_theft_hctsa`.

#### Calculating features

The core function in `theft` is `calculate_features` which takes the following arguments:

* `data`---a `tbl_ts` containing the time series data
* `feature_set`---character or vector of characters denoting the set of time-series features to calculate. Can be one or more of `"catch22"`, `"feasts"`, `"tsfeatures"`, `"tsfresh"`, `"tsfel"`, `"kats"`, `"quantiles"`, or `"moments"`
* `features`---a named list containing a set of user-supplied functions to calculate on `data`. Each function should take a single argument which is the time series. Defaults to `NULL` for no manually-specified features. Each list entry must have a name as `calculate_features` looks for these to name the features. If you don't want to use the existing feature sets and only compute those passed to `features`, set `feature_set = NULL`
* `catch24`---Boolean specifying whether to compute `catch24` in addition to `catch22` if `catch22` is one of the feature sets selected. Defaults to `FALSE`
* `tsfresh_cleanup`---Boolean specifying whether to use the in-built `tsfresh` relevant feature filter or not. Defaults to `FALSE`
* `use_compengine`---Boolean specifying whether to use the `"compengine"` features in `tsfeatures`. Defaults to `FALSE` to provide immense computational efficiency benefits
* `seed`---integer denoting a fixed number for R's random number generator to ensure reproducibility. Defaults to `123`
* `z_score`---Boolean specifying whether to *z*-score the time-series before computing features. Defaults to `FALSE`
* `n_jobs`---integer denoting the number of parallel processes to use if `"tsfresh"` or `"tsfel"` are specified in `feature_set`. Defaults to `0` for no parallelisation
* `warn`---Boolean specifying whether to produce warnings from feature set packages. Defaults to `TRUE`

Here is an example with the `catch22` set:
  
```{r, message = FALSE, warning = FALSE}
feature_matrix <- calculate_features(data = simData, feature_set = "catch22")
head(feature_matrix)
```

Note that `data` must be a `tsibble::tbl_ts` object, which has specified `key` (i.e., identifying) and `index` (i.e., time) variables. `theft` treats the first variable in the `key` as the ID variable and the second as the grouping variable (if there is one). Any other key variables will be ignored by `theft`.

You can also supply your own named list of functions to compute as time-series features. Below is an example with mean and standard deviation. Note that the list *must* be named as `theft` uses the list element names to label the time-series features internally. Note that if you don't want to use any of the existing feature sets in `theft` and only calculate the features you supply to `features`, just set `feature_set = NULL`.

```{r, message = FALSE, warning = FALSE}
feature_matrix2 <- calculate_features(data = simData, 
                                      feature_set = NULL,
                                      features = list("mean" = mean, "sd" = sd))

head(feature_matrix2)
```

### Comparison of feature sets

For a detailed comparison of the six feature sets, see [this paper](https://ieeexplore.ieee.org/document/9679937) for a detailed review^[T. Henderson and B. D. Fulcher, "An Empirical Evaluation of Time-Series Feature Sets," 2021 International Conference on Data Mining Workshops (ICDMW), 2021, pp. 1032-1038, doi: 10.1109/ICDMW53433.2021.00134.].

## Reading and processing hctsa-formatted files

As `theft` is based on the foundations laid by [`hctsa`](https://github.com/benfulcher/hctsa), there is also functionality for reading in `hctsa`-formatted Matlab files and automatically processing them into tidy dataframes ready for feature extraction in `theft`. The `process_hctsa_file` function takes a string filepath to the Matlab file and does all the work for you, returning a dataframe with naming conventions consistent with other `theft` functionality. As per `hctsa` specifications for [Input File Format 1](https://time-series-features.gitbook.io/hctsa-manual/installing-and-using-hctsa/calculating/input_files#input-file-format-1-.mat-file), this file should have 3 variables with the following exact names: `timeSeriesData`, `labels`, and `keywords`. The filepath can be a local drive path or a URL.

## Analysing, interpreting, and visualising time-series features

Please see the companion package [`theftdlc`](https://github.com/hendersontrent/theftdlc) ('`theft` downloadable content') for a large suite of functions that are designed to work on top of `theft`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_hctsa_file.R
\name{process_hctsa_file}
\alias{process_hctsa_file}
\title{Load in hctsa formatted MATLAB files of time series data into a tidy format ready for feature extraction}
\usage{
process_hctsa_file(data)
}
\arguments{
\item{data}{\code{string} specifying the filepath to the \code{MATLAB} file to parse}
}
\value{
an object of class \code{data.frame} in tidy format
}
\description{
Load in hctsa formatted MATLAB files of time series data into a tidy format ready for feature extraction
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/init_theft.R
\name{init_theft_kats}
\alias{init_theft_kats}
\title{Communicate to R the Python virtual environment containing kats only}
\usage{
init_theft_kats(venv)
}
\arguments{
\item{venv}{\code{character} specifying the name of the to the Python virtual environment where \code{"kats"} is installed}
}
\value{
no return value; called for side effects
}
\description{
Communicate to R the Python virtual environment containing kats only
}
\examples{
\dontrun{
install_python_pkgs("theft-test")
init_theft_kats("theft-test")
}

}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{feature_list}
\alias{feature_list}
\title{All features available in theft in tidy format}
\format{
A tidy data frame with 2 variables:
\describe{
  \item{feature_set}{Name of the set the feature is from}
  \item{feature}{Name of the feature}
}
}
\usage{
feature_list
}
\description{
The variables include:
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_vector_quality.R
\name{check_vector_quality}
\alias{check_vector_quality}
\title{Check for presence of NAs and non-numerics in a vector}
\usage{
check_vector_quality(x)
}
\arguments{
\item{x}{input \code{vector}}
}
\value{
\code{Boolean} of whether the data is good to extract features on or not
}
\description{
Check for presence of NAs and non-numerics in a vector
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{simData}
\alias{simData}
\title{Sample of randomly-generated time series to produce function tests and vignettes}
\format{
A tidy tsibble with 4 variables:
\describe{
  \item{id}{Unique identifier for the time series}
  \item{timepoint}{Time index}
  \item{values}{Value}
  \item{process}{Group label for the type of time series}
}
}
\usage{
simData
}
\description{
The variables include:
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_features.R
\name{calculate_features}
\alias{calculate_features}
\title{Compute features on an input time series dataset}
\usage{
calculate_features(
  data,
  feature_set = c("catch22", "feasts", "tsfeatures", "kats", "tsfresh", "tsfel", "hctsa",
    "quantiles", "moments"),
  features = NULL,
  catch24 = FALSE,
  tsfresh_cleanup = FALSE,
  use_compengine = FALSE,
  seed = 123,
  z_score = FALSE,
  n_jobs = 0,
  warn = TRUE
)
}
\arguments{
\item{data}{\code{tbl_ts} containing the time series data}

\item{feature_set}{\code{character} or \code{vector} of \code{character} denoting the set of time-series features to calculate. Can be one of \code{"catch22"}, \code{"feasts"}, \code{"tsfeatures"}, \code{"tsfresh"}, \code{"tsfel"}, \code{"kats"}, \code{"quantiles"}, and or \code{"moments"}}

\item{features}{named \code{list} containing a set of user-supplied functions to calculate on \code{data}. Each function should take a single argument which is the time series. Defaults to \code{NULL} for no manually-specified features. Each list entry must have a name as \code{calculate_features} looks for these to name the features. If you don't want to use the existing feature sets and only compute those passed to \code{features}, set \code{feature_set = NULL}}

\item{catch24}{\code{Boolean} specifying whether to compute \code{catch24} in addition to \code{catch22} if \code{catch22} is one of the feature sets selected. Defaults to \code{FALSE}}

\item{tsfresh_cleanup}{\code{Boolean} specifying whether to use the in-built \code{tsfresh} relevant feature filter or not. Defaults to \code{FALSE}}

\item{use_compengine}{\code{Boolean} specifying whether to use the \code{"compengine"} features in \code{tsfeatures}. Defaults to \code{FALSE} to provide immense computational efficiency benefits}

\item{seed}{\code{integer} denoting a fixed number for R's random number generator to ensure reproducibility. Defaults to \code{123}}

\item{z_score}{\code{Boolean} specifying whether to z-score the time-series before computing features. Defaults to \code{FALSE}}

\item{n_jobs}{\code{integer} denoting the number of parallel processes to use if \code{"tsfresh"} or \code{"tsfel"} are specified in \code{"feature_set"}. Defaults to \code{0} for no parallelisation}

\item{warn}{\code{Boolean} specifying whether to produce warnings from feature set packages. Defaults to \code{TRUE}}
}
\value{
object of class \code{feature_calculations} that contains the summary statistics for each feature
}
\description{
Compute features on an input time series dataset
}
\examples{
featMat <- calculate_features(data = simData, 
  feature_set = "catch22")

}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theft.R
\name{theft}
\alias{theft}
\title{Tools for Handling Extraction of Features from Time-series}
\description{
Tools for Handling Extraction of Features from Time-series
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moments.R
\name{moments}
\alias{moments}
\title{Calculate a basic set of the four moments of the distribution}
\usage{
moments(y)
}
\arguments{
\item{y}{\code{numeric} vector of values}
}
\value{
\code{data.frame} of results
}
\description{
Calculate a basic set of the four moments of the distribution
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moment_calcs.R
\name{kurtosis}
\alias{kurtosis}
\title{Calculate a kurtosis of a vector}
\usage{
kurtosis(y)
}
\arguments{
\item{y}{\code{numeric} vector of values}
}
\value{
\code{numeric} scalar of kurtosis
}
\description{
Calculate a kurtosis of a vector
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/init_theft.R
\name{init_theft}
\alias{init_theft}
\title{Communicate to R the Python virtual environment containing the relevant libraries for calculating features}
\usage{
init_theft(venv, hctsa = FALSE)
}
\arguments{
\item{venv}{\code{character} specifying the name of the to the Python virtual environment where \code{"tsfresh"}, \code{"TSFEL"}, and/or \code{"Kats"} are installed}

\item{hctsa}{\code{Boolean} denoting whether pyhctsa was installed or not. Defaults to \code{FALSE}}
}
\value{
no return value; called for side effects
}
\description{
Communicate to R the Python virtual environment containing the relevant libraries for calculating features
}
\examples{
\dontrun{
install_python_pkgs("theft-test")
init_theft("theft-test")
}

}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moment_calcs.R
\name{skewness}
\alias{skewness}
\title{Calculate a skewness of a vector}
\usage{
skewness(y)
}
\arguments{
\item{y}{\code{numeric} vector of values}
}
\value{
\code{numeric} scalar of skewness
}
\description{
Calculate a skewness of a vector
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quantiles.R
\name{quantiles}
\alias{quantiles}
\title{Calculate a basic set of quantiles for an input time-series vector}
\usage{
quantiles(y, quantiles = seq(0, 1, by = 0.01))
}
\arguments{
\item{y}{\code{numeric} vector of values}

\item{quantiles}{\code{numeric} vector of quantiles to calculate. Defaults to \code{seq(0, 1, by = 0.01)}}
}
\value{
\code{data.frame} of results
}
\description{
Calculate a basic set of quantiles for an input time-series vector
}
\author{
Trent Henderson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/installs.R
\name{install_kats}
\alias{install_kats}
\title{Download and install Kats from Python into a new virtual environment}
\usage{
install_kats(venv, python)
}
\arguments{
\item{venv}{\code{character} specifying the name of the new virtual environment to create}

\item{python}{\code{character} specifying the filepath to the Python interpreter to use. Python 3.10 is recommended}
}
\value{
no return value; called for side effects
}
\description{
Download and install Kats from Python into a new virtual environment
}
\examples{
\dontrun{
install_kats("theft-test", "/usr/local/bin/python3.10")
}

}
\author{
Trent Henderson
}
